@@ -34,6 +34,7 @@ import mr.shared_types
 import mr.constants
 import mr.handlers.scope
 import mr.utility
+import mr.log
 
 _logger = logging.getLogger(__name__)
 _flow_logger = _logger.getChild('flow')
@@ -484,7 +485,8 @@ class _StepProcessor(object):
         invocation = message_parameters.invocation
         workflow = message_parameters.workflow
         
-        _logger.debug("Processing MAP: [%s]", invocation)
+        _logger.debug("Processing MAP: [%s] [%s]", 
+                      invocation, invocation.created_timestamp)
 
         try:
             ## Call the handler.
@@ -554,14 +556,31 @@ class _StepProcessor(object):
             _logger.exception("Exception while processing MAP under request: "
                               "%s", request)
 
-# TODO(dustin): We might have to remove the chain of invocations, on error.
             invocation.error = traceback.format_exc()
             invocation.save()
 
+            # Formally mark the request as failed but finished. In the event 
+            # that request-cleanup is disabled, forensics will be intact.
+
             request.failed_invocation_id = invocation.invocation_id
-            request.done = True
+            request.is_done = True
             request.save()
 
+            # Send notification.
+
+            notify = mr.log.get_notify()
+            notify.exception("Mapper invocation [%s] under request [%s] "
+                             "failed. HANDLER=[%s]", 
+                             invocation.invocation_id, request.request_id, 
+                             step.map_handler_name)
+
+            # Schedule the request for destruction.
+
+            wm = mr.workflow_manager.get_wm()
+            managed_workflow = wm.get(workflow.workflow_name)
+
+            managed_workflow.cleanup_queue.add_request(request)
+
             raise
 
     def __decrement_invocation(self, workflow, invocation):
@@ -644,14 +663,31 @@ class _StepProcessor(object):
             _logger.exception("Exception while processing REDUCE under "
                               "request: %s", request)
 
-# TODO(dustin): We might have to remove the chain of invocations, on error.
+            # Formally mark the request as failed but finished. In the event 
+            # that request-cleanup is disabled, forensics will be intact.
+
             reduce_invocation.error = traceback.format_exc()
             reduce_invocation.save()
 
             request.failed_invocation_id = reduce_invocation.invocation_id
-            request.done = True
+            request.is_done = True
             request.save()
 
+            # Send notification.
+
+            notify = mr.log.get_notify()
+            notify.exception("Reducer invocation [%s] under request [%s] "
+                             "failed. HANDLER=[%s]", 
+                             reduce_invocation.invocation_id, 
+                             request.request_id, step.reduce_handler_name)
+
+            # Schedule the request for destruction.
+
+            wm = mr.workflow_manager.get_wm()
+            managed_workflow = wm.get(workflow.workflow_name)
+
+            managed_workflow.cleanup_queue.add_request(request)
+
             raise
 
     def __handle_mapped_mapping_reduce(self, message_parameters, step, 
@@ -704,7 +740,7 @@ class _StepProcessor(object):
                 # that we branched to.
 
                 _flow_logger.debug("  Reading constituent POST-REDUCE result "
-                                   "for parent [%s] by [%s] from: [%s]", 
+                                   "under parent [%s] for reducer [%s] from mapper: [%s]", 
                                    map_invocation, 
                                    message_parameters.invocation, 
                                    mapped_invocation)
@@ -737,6 +773,21 @@ class _StepProcessor(object):
 
         results_gen = (data['p'] for data in results_gen)
 
+        # Now group ("merge") the values for common keys. Since the individual 
+        # reducers and mappers can't be required to yield pairs sorted by key, 
+        # this will have a high [memory] cost for large sets, and requires us 
+        # to run the generators.
+
+        grouped_results = {}
+        for (k, v) in results_gen:
+            try:
+                grouped_results[k].append(v)
+            except KeyError:
+                grouped_results[k] = [v]
+
+        # We're not in Python3, so we have to use <dict>.iteritems().
+        grouped_results_gen = grouped_results.iteritems()
+
         # Disable session writes because there's no purpose by the time we get 
         # to the reducer.
         handler_ctx = self.__get_handler_context(
@@ -745,7 +796,7 @@ class _StepProcessor(object):
                         allow_session_writes=False)
 
         handler_arguments = {
-            'results': results_gen,
+            'results': grouped_results_gen,
             'ctx': handler_ctx,
         }
 
@@ -945,12 +996,36 @@ class _StepProcessor(object):
         else:
             # We've reduced our way back up to the original request.
 
-            _logger.debug("No further parents. Marking request as "
-                          "complete: [%s]", request.request_id)
+            _logger.debug("No further parents on request: %s", request)
 
-            request.done = True
+            # If we're a non-blocking request, This will be the last 
+            # opportunity to handle the result before sending the request into 
+            # oblivion.
+            if request.is_blocking is False:
+                _logger.debug("Writing result for non-blocking request: %s", 
+                              request)
+
+                rr = get_request_receiver()
+                rr.render_result(request)
+
+            _logger.debug("Marking request as complete: [%s]", 
+                          request.request_id)
+
+            request.is_done = True
             request.save()
 
+            # We allow for the result to be written into the request response, 
+            # and *then* cleanup. However, if the request was non-blocking, 
+            # we'll queue it for cleanup, now.
+            if request.is_blocking is False:
+                _logger.debug("Request is non-blocking, so we'll clean it up "
+                              "immediately: %s", request)
+
+                wm = mr.workflow_manager.get_wm()
+                managed_workflow = wm.get(workflow.workflow_name)
+
+                managed_workflow.cleanup_queue.add_request(request)
+
 _sp = _StepProcessor()
 
 def get_step_processor():
@@ -968,29 +1043,31 @@ class _RequestReceiver(object):
 
         self.__result_writer = result_writer_cls()
 
-    def __push_request(self, message_parameters):
+    def push_request(self, message_parameters):
         pusher = _get_pusher()
         pusher.queue_initial_map_step_from_parameters(message_parameters)
 
-    def __block_for_result(self, message_parameters):
-        request = message_parameters.request
+    def render_result(self, request):
+        """This is either called after a request has been posted (by the 
+        request handler, in a blocking request), or by the code that actually
+        posts the result (in a non-blocking request)."""
 
-        _logger.debug("Blocking on result for request: [%s]", 
-                      request.request_id)
+        workflow = mr.models.kv.workflow.get(request.workflow_name)
 
-        # The object is replaced with a newer one, when a change happens.
-        request = request.wait_for_change()
+        invocation = mr.models.kv.invocation.get(
+                        workflow, 
+                        request.invocation_id)
 
         _logger.debug("Reading result from: [%s] [%s]",
-                      message_parameters.invocation,
-                      message_parameters.invocation.direction)
+                      invocation,
+                      invocation.direction)
 
         _flow_logger.debug("  Reading POST-REDUCE dataset as final result: "
-                           "[%s]", message_parameters.invocation)
+                           "[%s]", invocation)
 
         dq = mr.models.kv.queues.dataset.DatasetQueue(
-                message_parameters.workflow, 
-                message_parameters.invocation, 
+                workflow, 
+                invocation, 
                 mr.models.kv.queues.dataset.DT_POST_REDUCE)
 
         result_pair_gen = (d['p'] for d in dq.list_data())
@@ -1000,12 +1077,37 @@ class _RequestReceiver(object):
             _logger.debug("Result to return for request:\n%s", 
                           pprint.pformat(result_pair_gen))
 
-        return self.__result_writer.get_response_tokens(
-                request.request_id, 
-                result_pair_gen)
+        result = self.__result_writer.render(
+                    request.request_id, 
+                    result_pair_gen)
+
+        # The result-writer can generate information that will be returned in 
+        # the request-response. However, this doesn't make sense if the request
+        # is asynchronous.
+        if result is not None and request.is_blocking is False:
+            raise ValueError("A result-writer can not return a value to a non-"
+                             "blocking request.")
+
+        return result
+
+    def block_for_result(self, message_parameters):
+        request = message_parameters.request
+
+        _logger.debug("Blocking on result for request: [%s]", request)
+
+        # The object is replaced with a newer one, when a change happens.
+        request = request.wait_for_change()
+
+        # We handle a result separately for a blocking versus a non-blocking 
+        # request, because we want to allow for a "result writer" that returns 
+        # the result in the request-response.
+
+        _logger.debug("Result is ready for blocking request: [%s]", request)
+
+        return self.render_result(request)
 
     def package_request(self, workflow, job, step, handler, arguments, 
-                        context):
+                        context, is_blocking=False):
         """Prepare an incoming request to be processed."""
 
         invocation = mr.models.kv.invocation.Invocation(
@@ -1036,15 +1138,11 @@ class _RequestReceiver(object):
                     workflow_name=workflow.workflow_name,
                     job_name=job.job_name, 
                     invocation_id=invocation.invocation_id,
-                    context=context)
+                    context=context,
+                    is_blocking=is_blocking)
 
         request.save()
 
-# TODO(dustin): The invocation and request are being saved here (confirmed), 
-#               but not found from the queue message-handlers if the current 
-#               request terminates too soon (when we're running 
-#               asynchronously).
-
         _logger.debug("Received request: [%s]", request)
 
         message_parameters = mr.shared_types.QUEUE_MESSAGE_PARAMETERS_CLS(
@@ -1057,15 +1155,6 @@ class _RequestReceiver(object):
 
         return message_parameters
 
-    def process_request(self, message_parameters, block=True):
-        self.__push_request(message_parameters)
-
-        if block is True:
-            r = self.__block_for_result(message_parameters)
-            return r
-        else:
-            return None
-
 _request_receiver = None
 def get_request_receiver():
     global _request_receiver
